package archive

import (
	"compress/gzip"
	"fmt"
	"io"
	"net/http"
	"strings"
	"time"

	"golang.org/x/crypto/openpgp/packet"

	"github.com/beevik/etree"
	"github.com/canonical/chisel/internal/cache"
	"github.com/canonical/chisel/internal/control"
	"github.com/canonical/chisel/internal/rpm"
)

type Archive interface {
	Options() *Options
	Fetch(pkg string) (io.ReadSeekCloser, *PackageInfo, error)
	Exists(pkg string) bool
	Info(pkg string) (*PackageInfo, error)
}

type PackageInfo struct {
	Name    string
	Version string
	Arch    string
	SHA256  string
}

type Options struct {
	Label      string
	Version    string
	Arch       string
	Suites     []string
	Components []string
	Pro        string
	CacheDir   string
	PubKeys    []*packet.PublicKey
}

func Open(options *Options) (Archive, error) {
	var err error
	if options.Arch == "" {
		options.Arch, err = rpm.InferArch()
	} else {
		err = rpm.ValidateArch(options.Arch)
	}
	if err != nil {
		return nil, err
	}
	return openTencentOS(options)
}

type fetchFlags uint

const (
	fetchBulk    fetchFlags = 1 << iota
	fetchDefault fetchFlags = 0
)

var httpClient = &http.Client{
	Timeout: 30 * time.Second,
}

var httpDo = httpClient.Do

var bulkClient = &http.Client{
	Timeout: 5 * time.Minute,
}

var bulkDo = bulkClient.Do

type tencentosArchive struct {
	options Options
	indexes []*tencentosIndex
	cache   *cache.Cache
	pubKeys []*packet.PublicKey
	baseURL string
	creds   *credentials
}

type tencentosIndex struct {
	label     string
	version   string
	arch      string
	suite     string
	component string
	release   control.Section
	xmldata   *etree.Document
	archive   *tencentosArchive
}

func (a *tencentosArchive) Options() *Options {
	return &a.options
}

func (a *tencentosArchive) Exists(pkg string) bool {
	_, _, err := a.selectPackage(pkg)
	return err == nil
}

func (a *tencentosArchive) selectPackage(pkgName string) (*etree.Element, *tencentosIndex, error) {
	var selectedVersion string
	var selectedPkg *etree.Element
	var selectedIndex *tencentosIndex
	for _, index := range a.indexes {
		if index.component == "primary" && index.xmldata != nil {
			metadata := index.xmldata.SelectElement("metadata")
			if metadata == nil {
				return nil, nil, fmt.Errorf("Could not find metadata element")
			}
			for _, pkg := range metadata.SelectElements("package") {
				if name := pkg.SelectElement("name"); name != nil {
					if name.Text() == pkgName {
						pkgVer := pkg.SelectElement("version")
						epoch := pkgVer.SelectAttr("epoch").Value
						ver := pkgVer.SelectAttr("ver").Value
						rel := pkgVer.SelectAttr("rel").Value

						version := fmt.Sprintf("%s:%s-%s", epoch, ver, rel)
						if selectedVersion == "" || rpm.CompareVersions(selectedVersion, version) < 0 {
							selectedVersion = version
							selectedPkg = pkg
							selectedIndex = index
						}
					}
				}
			}

		}
	}
	if selectedVersion == "" {
		return nil, nil, fmt.Errorf("cannot find package %q in archive", pkgName)
	}
	return selectedPkg, selectedIndex, nil
}

func (a *tencentosArchive) Fetch(pkg string) (io.ReadSeekCloser, *PackageInfo, error) {
	pkgElement, index, err := a.selectPackage(pkg)
	if err != nil {
		return nil, nil, err
	}
	filename := pkgElement.SelectElement("location").SelectAttr("href").Value
	checksum := pkgElement.SelectElement("checksum").Text()
	// suffix := fmt.Sprintf("%s/%s/Packages/%s", index.suite, index.arch, filename)
	// if a.options.Label == "tencentos-epol" {
	suffix := fmt.Sprintf("%s/%s/os/%s", index.suite, index.arch, filename)
	// }
	logf("Fetching %s...", suffix)
	reader, err := index.fetch(suffix, checksum, fetchBulk)
	if err != nil {
		return nil, nil, err
	}
	info := etreePackageInfo(pkgElement)
	return reader, info, nil
}

func (a *tencentosArchive) Info(pkg string) (*PackageInfo, error) {
	pkgElement, _, err := a.selectPackage(pkg)
	if err != nil {
		return nil, err
	}
	info := etreePackageInfo(pkgElement)
	return info, nil
}

const tencentosURL = "https://mirrors.tencent.com/tlinux/"
const tencentosEpolURL = "http://mirrors.tencent.com/epol/"

const (
	ProFIPS        = "fips"
	ProFIPSUpdates = "fips-updates"
	ProApps        = "esm-apps"
	ProInfra       = "esm-infra"
)

var proArchiveInfo = map[string]struct {
	BaseURL, Label string
}{
	ProFIPS: {
		BaseURL: "https://esm.ubuntu.com/fips/ubuntu/",
		Label:   "UbuntuFIPS",
	},
	ProFIPSUpdates: {
		BaseURL: "https://esm.ubuntu.com/fips-updates/ubuntu/",
		Label:   "UbuntuFIPSUpdates",
	},
	ProApps: {
		BaseURL: "https://esm.ubuntu.com/apps/ubuntu/",
		Label:   "UbuntuESMApps",
	},
	ProInfra: {
		BaseURL: "https://esm.ubuntu.com/infra/ubuntu/",
		Label:   "UbuntuESM",
	},
}

func archiveURL(label, pro, arch string) (string, *credentials, error) {
	if pro != "" {
		archiveInfo, ok := proArchiveInfo[pro]
		if !ok {
			return "", nil, fmt.Errorf("invalid pro value: %q", pro)
		}
		url := archiveInfo.BaseURL
		creds, err := findCredentials(url)
		if err != nil {
			return "", nil, err
		}
		return url, creds, nil
	}
	if label == "tencentos-epol" {
		return tencentosEpolURL, nil, nil
	}
	return tencentosURL, nil, nil
}

func openTencentOS(options *Options) (Archive, error) {
	if len(options.Components) == 0 {
		return nil, fmt.Errorf("archive options missing components")
	}
	if len(options.Suites) == 0 {
		return nil, fmt.Errorf("archive options missing suites")
	}
	if len(options.Version) == 0 {
		return nil, fmt.Errorf("archive options missing version")
	}

	baseURL, creds, err := archiveURL(options.Label, options.Pro, options.Arch)
	if err != nil {
		return nil, err
	}

	archive := &tencentosArchive{
		options: *options,
		cache: &cache.Cache{
			Dir: options.CacheDir,
		},
		pubKeys: options.PubKeys,
		baseURL: baseURL,
		creds:   creds,
	}

	for _, suite := range options.Suites {
		for _, component := range options.Components {
			index := &tencentosIndex{
				label:     options.Label,
				version:   options.Version,
				arch:      options.Arch,
				suite:     suite,
				component: component,
				archive:   archive,
			}
			//获取repodata里对应xml的内容
			err := index.fetchIndex()
			if err != nil {
				return nil, err
			}
			archive.indexes = append(archive.indexes, index)
		}
	}

	return archive, nil
}

func extractRepomd(component string, baseURL string) (string, string, error) {
	resp, err := http.Get(baseURL)
	if err != nil {
		return "", "", fmt.Errorf("HTTP GET error: %v", err)
	}
	defer resp.Body.Close()

	doc := etree.NewDocument()
	if _, err := doc.ReadFrom(resp.Body); err != nil {
		return "", "", fmt.Errorf("failed to parse XML document: %v", err)
	}

	repomd := doc.SelectElement("repomd")
	if repomd == nil {
		return "", "", fmt.Errorf("cannot find repomd element")
	}

	location := ""
	openChecksum := ""

	for _, data := range repomd.SelectElements("data") {
		if data.SelectAttr("type").Value == "primary" {
			openChecksum = data.SelectElement("open-checksum").Text()
			location = data.SelectElement("location").SelectAttr("href").Value
			break
		}

	}

	if location == "" || openChecksum == "" {
		return "", "", fmt.Errorf("cannot find primary file in repomd")
	}

	return location, openChecksum, nil
}

func (index *tencentosIndex) fetchIndex() error {
	//解析repomd.xml， 获取primary.xml.gz的文件位置
	repomdURL := fmt.Sprintf("%s/%s/%s/%s/os/repodata/repomd.xml", tencentosURL, index.version, index.suite, index.arch)
	if index.label == "tencentos-epol" {
		repomdURL = fmt.Sprintf("%s/%s/%s/%s/os/repodata/repomd.xml", tencentosEpolURL, index.version, index.suite, index.arch)
	}
	logf("fetching %s", repomdURL)
	xmlFile, checksum, err := extractRepomd(index.component, repomdURL)
	if err != nil {
		return fmt.Errorf("Error getting file list: %v\n", err)
	}
	// xmlPath := fmt.Sprintf("%s/%s/Packages/%s", index.suite, index.arch, xmlFile)
	// if index.label == "tencentos-epol" {
	xmlPath := fmt.Sprintf("%s/%s/os/%s", index.suite, index.arch, xmlFile)
	// }
	logf("fetching %s", xmlPath)
	//读取primary.xml.gz的内容
	reader, err := index.fetch(xmlPath, checksum, fetchBulk)
	if err != nil {
		return err
	}

	defer reader.Close()
	// 创建新的XML文档
	doc := etree.NewDocument()

	// 从reader读取并解析XML内容
	if _, err := doc.ReadFrom(reader); err != nil {
		return fmt.Errorf("failed to parse XML document: %v", err)
	}

	// 保存解析后的XML文档到index结构体中
	index.xmldata = doc

	return nil
}

func (index *tencentosIndex) fetch(suffix, digest string, flags fetchFlags) (io.ReadSeekCloser, error) {
	reader, err := index.archive.cache.Open(digest)
	if err == nil {
		return reader, nil
	} else if err != cache.MissErr {
		return nil, err
	}

	baseURL, creds := index.archive.baseURL, index.archive.creds
	url := fmt.Sprintf("%s%s/%s", baseURL, index.version, suffix)
	req, err := http.NewRequest("GET", url, nil)
	if err != nil {
		return nil, fmt.Errorf("cannot create HTTP request: %v", err)
	}
	if creds != nil && !creds.Empty() {
		req.SetBasicAuth(creds.Username, creds.Password)
	}
	var resp *http.Response
	if flags&fetchBulk != 0 {
		resp, err = bulkDo(req)
	} else {
		resp, err = httpDo(req)
	}
	if err != nil {
		return nil, fmt.Errorf("cannot talk to archive: %v", err)
	}
	defer resp.Body.Close()

	switch resp.StatusCode {
	case 200:
		// ok
	case 401:
		return nil, fmt.Errorf("cannot fetch from %q: unauthorized", index.label)
	case 404:
		return nil, fmt.Errorf("cannot find archive data")
	default:
		return nil, fmt.Errorf("error from archive: %v", resp.Status)
	}

	body := resp.Body
	if strings.HasSuffix(suffix, ".gz") {
		reader, err := gzip.NewReader(body)
		if err != nil {
			return nil, fmt.Errorf("cannot decompress data: %v", err)
		}
		defer reader.Close()
		body = reader
	}

	writer := index.archive.cache.Create(digest)
	defer writer.Close()

	_, err = io.Copy(writer, body)
	if err == nil {
		err = writer.Close()
	}
	if err != nil {
		return nil, fmt.Errorf("cannot fetch from archive: %v", err)
	}

	return index.archive.cache.Open(writer.Digest())
}

func etreePackageInfo(pkg *etree.Element) *PackageInfo {

	pkgVer := pkg.SelectElement("version")
	epoch := pkgVer.SelectAttr("epoch").Value
	ver := pkgVer.SelectAttr("ver").Value
	rel := pkgVer.SelectAttr("rel").Value
	ver_rel := fmt.Sprintf("%s:%s-%s", epoch, ver, rel)
	if epoch == "0" {
		ver_rel = fmt.Sprintf("%s-%s", ver, rel)
	}

	return &PackageInfo{
		Name:    pkg.SelectElement("name").Text(),
		Version: ver_rel,
		Arch:    pkg.SelectElement("arch").Text(),
		SHA256:  pkg.SelectElement("checksum").Text(),
	}
}
